@@ -589,7 +589,7 @@ CAMLprim value ocaml_mad_decode_frame_float(value madf) {
 
 CAMLprim value ocaml_mad_decode_frame_float_ba(value madf) {
   CAMLparam1(madf);
-  CAMLlocal1(ret);
+  CAMLlocal2(ret, tmp);
   madfile_t *mf = Madfile_val(madf);
   float *data;
   int chans;
@@ -603,14 +603,14 @@ CAMLprim value ocaml_mad_decode_frame_float_ba(value madf) {
   ret = caml_alloc_tuple(chans);
 
   for (c = 0; c < chans; c++) {
+    tmp = caml_ba_alloc_dims(CAML_BA_C_LAYOUT | CAML_BA_MANAGED, 1, NULL,
+                             mf->synth.pcm.length);
+    data = Caml_ba_data_val(tmp);
     caml_release_runtime_system();
-    data = malloc(mf->synth.pcm.length * sizeof(float));
     for (i = 0; i < mf->synth.pcm.length; i++)
       data[i] = mad_f_todouble(mf->synth.pcm.samples[c][i]);
     caml_acquire_runtime_system();
-    Store_field(ret, c,
-                caml_ba_alloc_dims(CAML_BA_C_LAYOUT | CAML_BA_MANAGED, 1, data,
-                                   mf->synth.pcm.length));
+    Store_field(ret, c, tmp);
   }
 
   CAMLreturn(ret);
@@ -621,13 +621,26 @@ CAMLprim value ocaml_mad_skip_frame(value madf) {
   madfile_t *mf = Madfile_val(madf);
   int err;
 
+  if (mf->eof)
+    caml_raise_constant(*caml_named_value("mad_exn_end_of_stream"));
+
   do {
     mf_fill_buffer(mf);
-    err = mf_decode(mf, 1);
-  } while (!mf->eof && err == 1);
 
-  if (err == 1)
-    caml_raise_constant(*caml_named_value("mad_exn_end_of_stream"));
+    caml_enter_blocking_section();
+    err = mad_header_decode(&mf->frame.header, &mf->stream);
+    caml_leave_blocking_section();
+
+    if (err == 0) {
+      mad_timer_add(&mf->timer, mf->frame.header.duration);
+    } else {
+      if (!MAD_RECOVERABLE(mf->stream.error) &&
+          mf->stream.error != MAD_ERROR_BUFLEN) {
+        caml_raise_with_arg(*caml_named_value("mad_exn_mad_error"),
+                            caml_copy_string(mad_stream_errorstr(&mf->stream)));
+      }
+    }
+  } while (!mf->eof);
 
   CAMLreturn(Val_unit);
 }
